// Copyright (C) 2017-2020  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Package tracetest provides infrastructure for testing concurrent systems
// based on synchronous event tracing.
//
// A serial system can be verified by checking that its execution produces
// expected serial stream of events. But concurrent systems cannot be verified
// by exactly this way because events are only partly-ordered with respect to
// each other by causality or so called happens-before relation.
//
// However in a concurrent system one can decompose all events into serial
// streams in which events should be strictly ordered by causality with respect
// to each other. This decomposition in turn allows to verify that in every
// stream events happenned as expected.
//
// Verification of events for all streams can be done by one *sequential*
// process:
//
//	- if events A and B in different streams are unrelated to each other by
// 	  causality, the sequence of checks models a particular possible flow of
// 	  time. Notably since events are delivered synchronously and sender is
// 	  blocked until receiver/checker explicitly confirms event has been
// 	  processed, by checking either A then B, or B then A allows to check
// 	  for a particular race-condition.
//
// 	- if events A and B in different streams are related to each other by
// 	  causality (i.e. there is some happens-before relation for them) the
// 	  sequence of checking should represent that ordering relation.
//
// The package should be used as follows:	XXX rework
//
//	  XXX tracer -> trace collector?
//	- implement tracer that will be synchronously collecting events from
//	  execution of your program. This can be done with package
//	  lab.nexedi.com/kirr/go123/tracing or by other similar means.
//
//	  the tracer have to output events to dispatcher (see below).
//
//	- implement router that will be making decisions specific to your	XXX -> Streams
//	  particular testing scenario on to which stream an event should belong.
//
//	  the router will be consulted by dispatcher (see below) for its working.
//
//	- create Dispatcher. This is the central place where events are
//	  delivered from tracer and are further delivered in accordance to what
//	  router says.
//
//	- for every serial stream of events create synchronous delivery channel
//	  (Chan) and event Checker. XXX
//
//
// XXX more text describing how to use the package.
// XXX link to example.
//
//
// XXX say that checker will detect deadlock if there is no event or it or
// another comes at different trace channel.
//
// XXX (if tested system is serial only there is no need to use Dispatcher and
// routing - the collector can send output directly to the only Chan with
// only one EventChecker connected to it).
package tracetest

import (
	"errors"
	"flag"
	"fmt"
	"sort"
	"strings"
	"sync"
	"reflect"
	"runtime"
	"runtime/debug"
	"testing"
	"time"

	"github.com/kylelemons/godebug/pretty"
)

var (
	chatty   = flag.Bool("tracetest.v", false, "verbose: print events as they are sent on trace channels")
	deadTime = flag.Duration("tracetest.deadtime", 3*time.Second, "time after which no events activity is considered to be a deadlock")
)

// _Chan provides synchronous channel with additional property that send
// blocks until receiving side explicitly acknowledges message was received and
// processed.
//
// New channels must be created via _NewChan.
//
// It is safe to use _Chan from multiple goroutines simultaneously.
type _Chan interface {
	_ChanTx
	_ChanRx
	name() string // name of the channel
}

// _ChanTx represents "send-only" half of _Chan.
// It is similar to chan<- .
type _ChanTx interface {
	// Send sends event to a consumer and waits for ack.
	// if main testing goroutine detects any problem Send panics.
	Send(event interface{})

	// Close closes the sending side of the channel.
	Close()
}

// _ChanRx represents "receive-only" half of _Chan.
// It is similar to <-chan .
type _ChanRx interface {
	// Recv receives message from a producer.
	//
	// The consumer, after dealing with the message, must send back an ack.
	Recv() *Msg

	// _rxq returns raw channel that is serving _ChanRx.
	// it is used internally to use _ChanRx in select.
	_rxq() <-chan *Msg
}

// Msg represents message with 1 event sent over _Chan.
//
// The goroutine which sent the message will wait for Ack before continue.
type Msg struct {
	Event interface {}
	ack   chan<- error // nil on Ack; !nil on nak
}



// _chan implements _Chan.
type _chan struct {
	t     *T
	msgq  chan *Msg
	down  chan struct{} // becomes ready when closed
	_name string
}

// name implements _Chan.
func (ch *_chan) name() string {
	return ch._name
}

// Send implements _ChanTx.
func (ch *_chan) Send(event interface{}) {
	if *chatty {
		fmt.Printf("%s <- %T %v\n", ch.name(), event, event)
	}
	ack := make(chan error)
	select {
	case <-ch.down:
		ch.t.fatalfInNonMain("%s: send: channel was closed", ch.name())

	case ch.msgq <- &Msg{event, ack}:
		err := <-ack
		if err != nil {
			ch.t.fatalfInNonMain("%s: send: %s", ch.name(), err)
		}
	}
}

// Close implements _ChanTx.
func (ch *_chan) Close() {
	close(ch.down) // note - not .msgq
}

// Recv implements _ChanRx.
func (ch *_chan) Recv() *Msg {
	msg := <-ch.msgq
	return msg
}

// _rxq implements _ChanRx.
func (ch *_chan) _rxq() <-chan *Msg {
	return ch.msgq
}

// XXX -> Unpause? Cont? Continue?
// Ack acknowledges the event was processed and unblocks producer goroutine.
func (m *Msg) Ack() {
	m.ack <- nil
}

// XXX it should be called only by tracetest internals from under Fatal
func (m *Msg) nak(why string) {
	m.ack <- errors.New(why)
}

// _NewChan creates new _Chan channel.
func _NewChan(name string) _Chan {
	// XXX somehow avoid channels with duplicate names
	//     (only allow to create named channels from under dispatcher?)
	return &_chan{msgq: make(chan *Msg), down: make(chan struct{}), _name: name}
}


// ----------------------------------------

/*

// XXX -> tracetest.TSerial ? tracetest.TSeq?
// EventChecker is testing utility to verify that sequence of events coming
// from a single Chan is as expected.
type EventChecker struct {
	t  testing.TB
	in Chan // XXX -> ChanRx ?
	dispatch *EventDispatcher // XXX why here?
}

// NewEventChecker constructs new EventChecker that will retrieve events from
// `in` and use `t` for tests reporting.
//
// XXX -> dispatch.NewChecker() ?
// XXX kill dispatch arg in NewEventChecker (and set it only from dispatch.NewChecker)
func NewEventChecker(t testing.TB, dispatch *EventDispatcher, in Chan) *EventChecker {
	return &EventChecker{t: t, in: in, dispatch: dispatch}
}

// get1 gets 1 event in place and checks it has expected type
//
// if checks do not pass - fatal testing error is raised
func (evc *EventChecker) xget1(eventp interface{}) *Msg {
	evc.t.Helper()
	var msg *Msg

	select {
	case msg = <-evc.in._rxq(): // unwrapped Recv
		// ok

	case <-time.After(*deadTime):
		evc.deadlock(eventp)	  // XXX -> Dispatcher + allow to tune?
	}

	reventp := reflect.ValueOf(eventp)
	if reventp.Type().Elem() != reflect.TypeOf(msg.Event) {
		// msg.ack <- error
		evc.t.Fatalf("%s: expect: %s:  got %#v", evc.in.name(), reventp.Elem().Type(), msg.Event)
	}

	// *eventp = msg.Event
	reventp.Elem().Set(reflect.ValueOf(msg.Event))

	return msg
}

// expect1 asks checker to expect next event to be eventExpect (both type and value)
//
// if checks do not pass - fatal testing error is raised.
func (evc *EventChecker) expect1(eventExpect interface{}) *Msg {
	evc.t.Helper()

	reventExpect := reflect.ValueOf(eventExpect)

	reventp := reflect.New(reventExpect.Type())
	msg := evc.xget1(reventp.Interface())
	revent := reventp.Elem()

	if !reflect.DeepEqual(revent.Interface(), reventExpect.Interface()) {
		// msg.ack <- error
		evc.t.Fatalf("%s: expect: %s:\nwant: %v\nhave: %v\ndiff: %s\n",
			evc.in.name(),
			reventExpect.Type(), reventExpect, revent,
			pretty.Compare(reventExpect.Interface(), revent.Interface()))
	}

	return msg
}

// Expect asks checker to receive next event and verify it to be equal to expected.
//
// If check is successful ACK is sent back to event producer.
// If check does not pass - fatal testing error is raised.
func (evc *EventChecker) Expect(expected interface{}) {
	evc.t.Helper()

	msg := evc.expect1(expected)
	msg.Ack()
}

// ExpectNoACK asks checker to receive next event and verify it to be equal to
// expected without sending back ACK.
//
// No ACK is sent back to event producer - the caller becomes responsible to
// send ACK back by itself.
//
// If check does not pass - fatal testing error is raised.
func (evc *EventChecker) ExpectNoACK(expected interface{}) *Msg {
	evc.t.Helper()

	msg := evc.expect1(expected)
	return msg
}


// deadlock reports diagnostic when retrieving event from .in times out.
//
// Timing out on recv means there is a deadlock either if no event was sent at
// all, or some other event was sent to another channel/checker.
//
// Report the full picture - what was expected and what was sent where.
func (evc *EventChecker) deadlock(eventp interface{}) {
	evc.t.Helper()

	bad := fmt.Sprintf("%s: deadlock waiting for %T\n", evc.in.name(), eventp)
	type sendInfo struct{dst Chan; event interface{}}
	var sendv []sendInfo
	for _, dst := range evc.dispatch.streams {
		// check whether someone is sending on a dst without blocking.
		// if yes - report to sender there is a problem - so it can cancel its task.
		select {
		case msg := <-dst._rxq():
			sendv = append(sendv, sendInfo{dst, msg.Event})
			//msg.ack <- false

		default:
		}

		// XXX panic triggering disabled because if sender panics we have no chance to continue
		// TODO retest this

		// in any case close channel where future Sends may arrive so that will panic too.
		//close(dst.msgq)
	}

	// order channels by name
	sort.Slice(sendv, func(i, j int) bool {
		return strings.Compare(sendv[i].dst.name(), sendv[j].dst.name()) < 0
	})

	if len(sendv) == 0 {
		bad += fmt.Sprintf("noone is sending\n")
	} else {
		bad += fmt.Sprintf("there are %d pending sender(s) on other channel(s):\n", len(sendv))
		for _, __ := range sendv {
			bad += fmt.Sprintf("%s:\t%T %v\n", __.dst.name(), __.event, __.event)
		}
	}

	evc.t.Fatal(bad)
}

*/


// ----------------------------------------

// XXX -> Streams
/*
// EventRouter is the interface used for routing events to appropriate output Chan.
//
// It should be safe to use EventRouter from multiple goroutines simultaneously.
type EventRouter interface {
	// Route should return appropriate destination for event.
	Route(event interface{}) *Chan

	// AllRoutes should return all routing destinations.
	AllRoutes() []*Chan
}
*/

// // EventDispatcher dispatches events to appropriate Chan for checking
// // according to provided streams.
// type EventDispatcher struct {
// 	streams    map[/*name*/string]Chan	// XXX -> ChanTx? just str
// 	routeEvent func(event interface{}) Chan // XXX ----//----
// }

/*
// NewEventDispatcher creates new dispatcher and provides router to it.
func NewEventDispatcher(streams Streams) *EventDispatcher {
	// build streams as dict and make sure all streams are named uniquely
	streamDict := make(map[string]Chan)
	for _, ch := range streams.AllStreams() {
		name := ch.name()
		_, already := streamDict[name]
		if already {
			panicf("duplicate stream name %q", name)
		}
		streamDict[name] = ch
	}
	return &EventDispatcher{
		streams:    streamDict,
		routeEvent: streams.RouteEvent,
	}
}
*/

// // XXX -> tracetest.T ?
// // Dispatch dispatches event to appropriate output channel.
// //
// // It is safe to use Dispatch from multiple goroutines simultaneously.
// func (d *EventDispatcher) Dispatch(event interface{}) {
// 	ch := d.routeEvent(event)
// 
// 	// assert that ch is within AllStreams() return.
// 	if ch == nil {
// 		panicf("Event %#v routed to nil", event)
// 	}
// /* XXX deactivated for now because neo.tNewCluster creates NewEventDispatcher with initially empty router
// 	ch_, ok := d.streams[ch.name]
// 	if !ok {
// 		all := []string{}
// 		for name := range d.streams {
// 			all = append(all, name)
// 		}
// 		sort.Slice(all, func(i, j int) bool {
// 			return strings.Compare(all[i], all[j]) < 0
// 		})
// 		panicf("Event %#v routed to stream %q that is outside of AllStreams %q", event, ch.name, all)
// 	}
// 	if ch != ch_ {
// 		panicf("Event %#v routed to stream %q, but channel is different from what AllSteams[%q] originally reported", event, ch.name, ch.name)
// 	}
// */
// 
// 	// TODO it is possible to empirically detect here if a test incorrectly
// 	// decomposed its system into serial streams: consider unrelated to each
// 	// other events A and B are incorrectly routed to the same channel. It
// 	// could be so happening that the order of checks on the test side is
// 	// almost always correct and so the error is not visible. However
// 	//
// 	//	if we add delays to delivery of either A or B
// 	//	and test both combinations
// 	//
// 	// we will for sure detect the error as, if A and B are indeed
// 	// unrelated, one of the delay combination will result in events
// 	// delivered to test in different to what it expects order.
// 	//
// 	// the time for delay could be taken as follows:
// 	//
// 	//	- run the test without delay; collect δt between events on particular stream
// 	//	- take delay = max(δt)·10
// 	//
// 	// to make sure there is indeed no different orderings possible on the
// 	// stream, rerun the test N(event-on-stream) times, and during i'th run
// 	// delay i'th event.
// 	//
// 	// TODO See also on this topic:
// 	// http://www.1024cores.net/home/relacy-race-detector
// 	// http://www.1024cores.net/home/relacy-race-detector/rrd-introduction
// 
// 
// 	// TODO timeout: deadlock? (print all-in-flight events on timeout)
// 	// XXX  or better ^^^ to do on receiver side?
// 	//
// 	// XXX -> if deadlock detection is done on receiver side (so in
// 	// EventChecker) -> we don't need EventDispatcher at all?
// 	ch.Send(event)
// }


/*

// Streams represents decomposition for events in a system into serial streams.
type Streams interface {
	// AllStreams should return all streams of the system.
	// Streams must have unique names.
	AllStreams() []Chan

	// RouteEvent should return particular stream into which event should be routed.
	// Returned stream must be one of reported by AllStreams.
	RouteEvent(event interface{}) Chan
}

*/

// ------ vvv = ok ------

// _testing_TB is alias for testing.TB that is non-public when embedded into a struct.
type _testing_TB = testing.TB

// T is similar to testing.T and is passed by Verify to tested function.
//
// See top-level package documentation for details.
type T struct {
	_testing_TB

	mu             sync.Mutex
	streamTab      map[/*stream*/string]_Chan // set to nil on test shutdown
	routeEvent     func(event interface{}) (stream string)
	tracev         []eventTrace // record of events as they happen
	delayInjectTab map[/*stream*/string]*delayInjectState

	nakq []nak // naks queued to be sent after Fatal
}

// eventTrace keeps information about one event T received via RxEvent.
type eventTrace struct {
	t      time.Time   // monotonic
	stream string
	event  interface{}
}

// delayInjectState is used by delay-injector to find out for which event on a
// stream a delay should be injected.
type delayInjectState struct {
	seqno   int           // current sequence number of event on this stream.
	delayAt int           // event with `seqno == delayAt` will be delayed
	delayT  time.Duration // by delayT time.
}

type nak struct {
	msg *Msg
	why string
}

// XXX place; just use t.Cleanup instead?
func (t *T) queuenak(msg *Msg, why string) {
	t.nakq = append(t.nakq, nak{msg, why})
}


// XXX doc; naming
// XXX document routeEvent requirements
func (t *T) SetEventRouter(routeEvent func(event interface{}) (stream string)) {
	t.mu.Lock()
	defer t.mu.Unlock()

	if t.routeEvent != nil {
		panic("double call to SetEventRouter")
	}
	t.routeEvent = routeEvent
}

// XXX doc; naming
// XXX NewEvent? IncomingEvent? OnEvent?
// XXX pass events in not via t (via something dedicated)?
func (t *T) RxEvent(event interface{}) {
	t0 := time.Now()
	stream := ""
	t.mu.Lock()
	if t.routeEvent != nil {
		stream = t.routeEvent(event)
	}
	if stream == "" {
		stream = "default"
	}
	t.tracev = append(t.tracev, eventTrace{t0, stream, event})
	ch := t.chanForStream(stream)

	var delay time.Duration
	d, ok := t.delayInjectTab[stream]
	if ok {
		if d.seqno == d.delayAt {
			delay = d.delayT
		}
		d.seqno++
	}

	t.mu.Unlock()

	if ch == nil {
		t.fatalfInNonMain("%s: (pre)send: canceled (test failed)", stream)
	}

	// TODO it is possible to empirically detect here if a test incorrectly
	// decomposed its system into serial streams: consider unrelated to each
	// other events A and B are incorrectly routed to the same channel. It
	// could be so happening that the order of checks on the test side is
	// almost always correct and so the error is not visible. However
	//
	//	if we add delays to delivery of either A or B
	//	and test both combinations
	//
	// we will for sure detect the error as, if A and B are indeed
	// unrelated, one of the delay combination will result in events
	// delivered to test in different to what it expects order.
	//
	// the time for delay could be taken as follows:
	//
	//	- run the test without delay; collect δt between events on particular stream
	//	- take delay = max(δt)·10
	//
	// to make sure there is indeed no different orderings possible on the
	// stream, rerun the test N(event-on-stream) times, and during i'th run
	// delay i'th event.
	//
	// TODO See also on this topic:
	// http://www.1024cores.net/home/relacy-race-detector
	// http://www.1024cores.net/home/relacy-race-detector/rrd-introduction

	if delay != 0 {
		time.Sleep(delay)
	}

	ch.Send(event)
}

// chanForStream returns channel corresponding to stream.
// must be called under mu.
func (t *T) chanForStream(stream string) _Chan {
	if t.streamTab == nil {
		return nil // t is no longer operational after e.g. deadlock
	}

	ch, ok := t.streamTab[stream]
	if !ok {
		ch = _NewChan(stream)
		ch.(*_chan).t = t // XXX move into _NewChan
		t.streamTab[stream] = ch
	}
	return ch
}

// Expect receives next event on stream and verifies it to be equal to eventOK.
//
// If check is successful ACK is sent back to event producer.
// If check does not pass - fatal testing error is raised.
func (t *T) Expect(stream string, eventOK interface{}) {
	t.Helper()
	msg := t.expect1(stream, eventOK)
	msg.Ack()
}

// XXX ExpectNoACK
// XXX Recv
// XXX Select? (e.g. Select("a", "b") to fetch from either "a" or "b")

// expect1 receives next event on stream and verifies it to be equal to eventOK (both type and value).
//
// if checks do not pass - fatal testing error is raised.
func (t *T) expect1(stream string, eventExpect interface{}) *Msg {
	t.Helper()

	reventExpect := reflect.ValueOf(eventExpect)

	reventp := reflect.New(reventExpect.Type())
	msg := t.xget1(stream, reventp.Interface())
	revent := reventp.Elem()

	if !reflect.DeepEqual(revent.Interface(), reventExpect.Interface()) {
		t.queuenak(msg, "unexpected event data")
		t.Fatalf("%s: expect: %s:\nwant: %v\nhave: %v\ndiff:\n%s\n\n",
			stream,
			reventExpect.Type(), reventExpect, revent,
			pretty.Compare(reventExpect.Interface(), revent.Interface()))
	}

	return msg
}

// xget1 gets 1 event in place and checks it has expected type
//
// if checks do not pass - fatal testing error is raised
func (t *T) xget1(stream string, eventp interface{}) *Msg {
	t.Helper()

	t.mu.Lock()
	ch := t.chanForStream(stream)
	t.mu.Unlock()

	// XXX ch == nil -> no longer operational

	var msg *Msg

	select {
	case msg = <-ch._rxq(): // unwrapped Recv
		// ok

	case <-time.After(*deadTime):
		t.Fatalf("%s: recv: deadlock waiting for %T\n", stream, eventp)
	}

	reventp := reflect.ValueOf(eventp)
	if reventp.Type().Elem() != reflect.TypeOf(msg.Event) {
		t.queuenak(msg, "unexpected event type")
		t.Fatalf("%s: expect: %s:  got %T %v", ch.name(), reventp.Elem().Type(), msg.Event, msg.Event)
	}

	// *eventp = msg.Event
	reventp.Elem().Set(reflect.ValueOf(msg.Event))

	return msg
}

// fatalfInNonMain should be called for fatal cases in non-main goroutines instead of panic.
//
// we don't panic because it will stop the process and prevent the main
// goroutine to print detailed reason for e.g. deadlock or other error.
var fatalLogMu sync.Mutex
func (t *T) fatalfInNonMain(format string, argv ...interface{}) {
	t.Helper()

	// serialize fatal log+traceback printout, so that such printouts from
	// multiple goroutines do not get intermixed.
	fatalLogMu.Lock()
	defer fatalLogMu.Unlock()

	t.Logf(format, argv...)
	t.Logf("%s\n", debug.Stack())
	runtime.Goexit()
}

// T overrides FailNow/Fatal/Fatalf to also cancel all in-progress sends.
func (t *T) FailNow() {
	t.Helper()
	_ = t.closeStreamTab()
	t._testing_TB.FailNow()
}

func (t *T) Fatal(argv ...interface{}) {
	t.Helper()
	t.Log(argv...)
	t.FailNow()
}

func (t *T) Fatalf(format string, argv ...interface{}) {
	t.Helper()
	t.Logf(format, argv...)
	t.FailNow()
}

// closeStreamTab prints details about pending event on streamTab, naks them
// and closes all channels. It returns the number of naked messages.
func (t *T) closeStreamTab() (nnak int) {
	t.Helper()

	// mark streamTab no longer operational
	t.mu.Lock()
	streamTab := t.streamTab
	t.streamTab = nil
	t.mu.Unlock()

	if streamTab == nil {
		return // already closed
	}

	// print details about pending events and all streams
	type sendInfo struct{ch _Chan; msg *Msg}
	var sendv  []sendInfo // sends are pending here
	var quietv []_Chan    // this channels are quiet

	// order channels by name
	var streams []string
	for __ := range streamTab {
		streams = append(streams, __)
	}
	sort.Slice(streams, func(i, j int) bool {
		return strings.Compare(streams[i], streams[j]) < 0
	})

	for _, stream := range streams {
		ch := streamTab[stream]
		// check whether someone is sending on channels without blocking.
		select {
		case msg := <-ch._rxq():
			sendv = append(sendv, sendInfo{ch, msg})
		default:
			quietv = append(quietv, ch)
		}
	}

	pending := fmt.Sprintf("test shutdown: #streams: %d,  #(pending events): %d\n", len(streams), len(sendv))
	for _, __ := range sendv {
		pending += fmt.Sprintf("%s\t<- %T %v\n", __.ch.name(), __.msg.Event, __.msg.Event)
	}
	for _, ch := range quietv {
		pending += fmt.Sprintf("# %s\n", ch.name())
	}

	// log the details and nak senders that we received from.
	// nak them only after details printout, so that our text comes first,
	// and their "panics" don't get intermixed with it.
	t.Log(pending)
	for _, __ := range t.nakq {
		__.msg.nak(__.why)
		nnak++
	}
	t.nakq = nil
	for _, __ := range sendv {
		__.msg.nak("canceled (test failed)")
		nnak++
	}
	// in any case close channel where future Sends may arrive so that they will "panic" too.
	for _, ch := range streamTab {
		ch.Close()
	}

	return nnak
}



// TODO func Verify(system)
// system() ->:
// (Streams)
// - []Chan			where events will be emitted
// - Route(event) -> Chan	which event goes to which event stream (= channel)
//
// - main			func to run the system
// - testf			(serial) func to verify events that comes from channels

// Verify verifies a test system.
// XXX
// XXX try back -> testing.TB? (but TB does not have .Run)
func Verify(t *testing.T, f func(t *T)) {
	testf := func(t testing.TB, delayInjectTab map[string]*delayInjectState) *T {
		tT := &T{
			_testing_TB:    t,
			streamTab:      make(map[string]_Chan),
			delayInjectTab: delayInjectTab,
		}

		// verify in the end that no events are left unchecked / unconsumed,
		// e.g. sent to RxEvent, but not received. Nak them if they are and fail.
		//
		// NOTE this complements T.Fatal and friends, because a test might
		// think it completes successfully, but leaves unconsumed events behind it.
		defer func() {
			nnak := tT.closeStreamTab()
			if nnak != 0 {
				t.Fatal()
			}
		}()

		f(tT)
		return tT
	}

	tT0 := testf(t, nil)

	// now, if f succeeds, verify f with injected delays.
	// XXX explain with more text
	if tT0.Failed() {
		return
	}

	trace0 := tT0.tracev
	if len(trace0) < 2 {
		return
	}
	streams0 := streamsOfTrace(trace0)

	// sort trace0 by time just in case - events migth come from multiple
	// CPUs simultaneously, and so for close events they might be added to
	// tracev not in time order.
	sort.Slice(trace0, func(i, j int) bool {
		return trace0[i].t.Before(trace0[j].t)
	})

	// find out max(δt) in between events
	var δtMax time.Duration
	for i := 1; i < len(trace0); i++ {
		δt := trace0[i].t.Sub(trace0[i-1].t)
		if δt > δtMax {
			δtMax = δt
		}
	}

	// retest f with 10·δtMax delay injected at i'th event
	delayT    := 10*δtMax            // TODO make sure it < deadTime
	delayTmin := 10*time.Millisecond // make sure delayT ≥ 10ms
	if delayT < delayTmin {
		delayT = delayTmin
	}
	for i := 0; i < len(trace0); i++ {
		// stream and on-stream sequence number for i'th global event
		stream := trace0[i].stream
		istream := -1
		for j := 0; j <= i; j++ {
			if trace0[j].stream == stream {
				istream++
			}
		}

		t.Run(fmt.Sprintf("delay@%d(=%s:%d)", i, stream, istream), func(t *testing.T) {
			tT := testf(t, map[string]*delayInjectState{
				stream: &delayInjectState{
					delayAt: istream,
					delayT:  delayT,
				},
			})

			// verify that streams are the same from run to run
			if tT.Failed() {
				return
			}
			streams := streamsOfTrace(tT.tracev)
			if !reflect.DeepEqual(streams, streams0) {
				tT.Fatalf("streams are not the same as in the first run:\n"+
					  "first: %s\nnow:   %s\ndiff:\n%s\n\n",
					  streams0, streams, pretty.Compare(streams0, streams))
			}
		})
	}
}


// streamsOfTrace returns sorted list of all streams present in a trace.
func streamsOfTrace(tracev []eventTrace) []string {
	streams := make(map[string]struct{})
	for _, t := range tracev {
		streams[t.stream] = struct{}{}
	}
	streamv := []string{}
	for stream := range streams {
		streamv = append(streamv, stream)
	}
	sort.Strings(streamv)
	return streamv
}
